# -*- coding: UTF-8 -*-
'''
    nereid.logging

    Logging Support for the Application

    :copyright: (c) 2010-2012 by Openlabs Technologies & Consulting (P) Ltd.
    :license: GPLv3, see LICENSE for more details
'''

from flask.logging import create_logger
